/*      file     : ParSim/TParSimSlave.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TParSimSlave.C,v 1.14 2005/03/25 11:07:34 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <unistd.h>

#include <ParToolsConfig.h>
#include <TConfiguration.h>
#include <TParallel.h>

#include <TrafficDataConfig.h>
#include <TObjectManager.h>
#include <TDynamicBalancer.h>
#include <TTransferInfo.h>
#include <TDistributedRange.h>
#include <TNode.h>
#include <TEdge.h>
#include <TTrafficDataManager.h>
#include <TTrafficSimulator.h>
#include <TRouter.h>

#include "ParSimConfig.h"

#ifdef XWINDOWS
#  include "TParSimGraphics.h"
#endif

#include "TEdgeInfo.h"
#include "TParSimSlave.h"

// ***DEFINES************************************************************************* 

#define PARSIM_DEFAULT_DELAY        0L
#define PARSIM_DEFAULT_GRAPHICS     0L

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TApplicationSlave
// =================================================================================== 

TParSimSlave::TParSimSlave(int argc, char *argv[]) :

ParentClass(argc, argv)

{
    Router = new TRouter;
    TrafficDataManager = new TTrafficDataManager;
    Simulator = new TTrafficSimulator(&ApplicationNodes, &ApplicationEdges,
                                      &BoundaryEdges, &NonBoundaryEdges,
                                      NodeTimer, EdgeTimer);
#ifdef XWINDOWS    
    Graphics = 0;
#endif
}

// =================================================================================== 

TParSimSlave::~TParSimSlave(void)

{
#ifdef XWINDOWS
    DeleteObject(Graphics);
#endif
    DeleteObject(Router);
    DeleteObject(Simulator);
    DeleteObject(TrafficDataManager);
}

// =================================================================================== 

bool
TParSimSlave::Activate(void)

{
    if (!TApplicationSlave::Activate())
        return FALSE;
    if (!TrafficDataManager->Activate())
    {
        Error("Activate: cannot activate traffic data manager");
        return FALSE;
    }
    if (!Simulator->Activate())
    {
        Error("Activate: cannot activate traffic simulator");
        return FALSE;
    }
    if (!Router->Activate())
    {
        Error("Activate: cannot activate router");
        return FALSE;
    }
    Configuration->SetDefault(TConfigEntry(TRAFFIC_GROUP, TRAFFIC_NAME_STARTTIME, 
                                           TRAFFIC_DEFAULT_STARTTIME));
    Configuration->SetDefault(TConfigEntry(PARSIM_GROUP, PARSIM_NAME_DELAY, 
                                           PARSIM_DEFAULT_DELAY));
    TTimeStep::SetNrOfSubTimeSteps(2);
    return TRUE;
}

// =================================================================================== 

bool
TParSimSlave::ExecuteSubTimeStep(void)

{
    TTransferInfo TransferInfo;

    if (!ParentClass::ExecuteSubTimeStep())
        return FALSE;
    TransferInfo.TransferMode = TM_SimpleBoundaries;
    if (!ObjectManager->SendBoundaries(ApplicationTimeStep, &TransferInfo))
    {
        Error("ExecuteTimeStep: cannot send boundaries");
        return FALSE;
    }
    if (!Simulator->ExecuteSubTimeStep(ApplicationTimeStep, FALSE))
    {
        Error("ExecuteTimeStep: during simulation (boundary-independant portion)");
        return FALSE;
    }
    if (!ObjectManager->ReceiveBoundaries())
    {
        Error("ExecuteTimeStep: while waiting for boundaries");
        return FALSE;
    }
    if (!Simulator->ExecuteSubTimeStep(ApplicationTimeStep, TRUE))
    {
        Error("ExecuteTimeStep: during simulation (boundary-dependant portion)");
        return FALSE;
    }
    if (!Router->ExecuteSubTimeStep(ApplicationTimeStep + StartTimeStep))
    {
        Error("ExecuteTimeStep: while calling router");
        return FALSE;
    }
#ifdef XWINDOWS
    if (Graphics)
        if (!Graphics->ExecuteSubTimeStep(ApplicationTimeStep))
        {
            Error("ExecuteSubTimeStep: while handling graphics");
            return FALSE;
        }
#endif
    if (Delay)
        sleep(Delay);
    return TRUE;
}

// =================================================================================== 

TString
TParSimSlave::GetClassName(void) const

{
    return "TParSimSlave";
}

// =================================================================================== 

bool
TParSimSlave::IdleTime(void)

{
#ifdef XWINDOWS
    if (Graphics)
        if (!Graphics->ExecuteSubTimeStep(ApplicationTimeStep))
        {
            Error("ExecuteSubTimeStep: while handling graphics");
            return FALSE;
        }
#endif
    return ParentClass::IdleTime();    
}

// =================================================================================== 

bool
TParSimSlave::InitializeDataAfterReception(void)

{
    TNodeEnum         CNode;
    TEdgeEnum         CEdge;
    TDistributedRange ActiveRange;

    if (!ParentClass::InitializeDataAfterReception())
        return FALSE;
    CEdge.WalksForwardThrough(InitEdges);
    while(CEdge())
    {
        CEdge->GetActiveRange(Parallel->GetLocalCPNID(), &ActiveRange);
        if (!ActiveRange.IsActive())
        {
            Error("InitializeDataAfterReception: %s is inactive", 
                  (char *) CEdge->GetObjectName());
            return FALSE;
        }
        if (!Simulator->InitializeEdgeAfterReception((TEdge *) CEdge, &ActiveRange))
        {
            Error("cannot initialize %s after reception",
                  (char *) CEdge->GetObjectName());
            return FALSE;
        }       
        CEdge++;
    }
    CNode.WalksForwardThrough(InitNodes);
    while(CNode())
    {
        if (!Simulator->InitializeNodeAfterReception((TNode *) CNode))
        {
            Error("cannot initialize %s after reception", 
                  (char *) CNode->GetObjectName());
            return FALSE;
        }
        CNode++;
    }
    InitEdges.DeleteAll();
    InitNodes.DeleteAll();
    return TRUE;
}

// =================================================================================== 

bool
TParSimSlave::Run(void)

{
    if (!ParentClass::Run())
        return FALSE;
    StartTimeStep = (long) (3600 *
        Configuration->GetDouble(TRAFFIC_GROUP, TRAFFIC_NAME_STARTTIME));
    Delay = Configuration->GetLong(PARSIM_GROUP, PARSIM_NAME_DELAY);
    if (!TrafficDataManager->Run())
        return FALSE;
    if (!Simulator->Run(StartTimeStep))
    {
        Error("Run: cannot run traffic simulator");
        return FALSE;
    }
    if (!Router->Run(StartTimeStep))
    {
        Error("Run: cannot run router");
        return FALSE;
    }
#ifdef XWINDOWS
#ifdef XWINDOWS
    Configuration->SetDefault(TConfigEntry(PARSIM_GROUP, PARSIM_NAME_GRAPHICS,
                                           PARSIM_DEFAULT_GRAPHICS));
    if (Configuration->GetLong(PARSIM_GROUP, PARSIM_NAME_GRAPHICS))
    {
        ObjectManager->SetGlobalMigrationMonitor(DEST_PARENT);
        if (Parallel->IsMaster())
        {
            Graphics = new TParSimGraphics;
            if (!Graphics->Activate())
            {
                Error("Activate: cannot activate graphics");
                return FALSE;
            }
        }
    }
#endif
    if (Graphics)
    {
        if (!Graphics->Run())
        {
            Error("Run: cannot run graphics");
            return FALSE;
        }
        if (DynamicBalancer)            
            if (!DynamicBalancer->StartNeighborCommStat())
            {
                Error("Activate: cannot start neighbor comm statistics");
                return FALSE;
            }
    }
#endif
    if (!ObjectManager->Wait(Parallel->GetParentCPNID()))
    {
        Error("Run: while waiting for initial topology");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
