/*      file     : TrafficData/TIntersectionControl.C
        project  : PAMINA III
        author   : Marcus Rickert
        CVS      : $Id: TIntersectionControl.C,v 1.5 2005/03/25 12:28:51 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************

#include <math.h>

#include <TBaseStream.h>
#include <TParser.h>
#include <TTimeStep.h>
#include <TTransferInfo.h>
#include <TIncidenceInfo.h>
#include <TEdge.h>
#include <TPhase.h>
#include <TAsciiDataTypeReader.h>

#include "Transims.h"
#include "TIntersectionControl.h"

// ***DEFINES*************************************************************************

#define DEFAULT_OFFSET           (-1)
#define DEFAULT_GREEN_PHASE      60
#define DEFAULT_TRANSITION_PHASE 5

//! minimum phase length in seconds 
#define MIN_PHASE_LENGTH         5

//! maximum phase length in seconds
#define MAX_PHASE_LENGTH       120


// ***GLOBAL VARIABLES****************************************************************  

TClassInfo *TIntersectionControl::ClassInfo = 0;

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TParallelObject
// =================================================================================== 

TIntersectionControl::TIntersectionControl(TParallelID theParallelID) :

ParentClass(ClassInfo, theParallelID)

{
    _Init();
}

// ===================================================================================

TIntersectionControl::~TIntersectionControl(void)

{
    if (Phases)
    {
        int i;
        
        for (i = 0; i < NrOfPhases; i++)
            DeleteObject(Phases[i]);
        DeleteObjectArray(Phases);
    }
    DeleteObject(Random);
}

// ===================================================================================

bool
TIntersectionControl::Activate(TIncidenceInfo *theIncidenceInfo)

{
    IncidenceInfo = theIncidenceInfo;    
    if (!SortPhases())
        return FALSE;
    ComputeOffsets();

    //    Print("Activate");

    return TRUE;
}

// ===================================================================================

bool
TIntersectionControl::Allocate(void)
    
{
    int CPhase;

    if (Phases == 0 AND NrOfPhases > 0)
    {
        Phases = new TPhase * [NrOfPhases];
        for (CPhase = 0; CPhase < NrOfPhases; CPhase++)
            Phases[CPhase] = new TPhase;
    }
    return TRUE;
}

// ===================================================================================

void
TIntersectionControl::ComputeOffsets(void)

{
    int CPhase;
    
    for (CPhase = 0; CPhase < NrOfPhases; CPhase++)
        Phases[CPhase]->ComputeOffsets(CPhase & 1);
}

// ===================================================================================

bool
TIntersectionControl::Decode(TBaseStream *aStream, TTransferInfo *TransferInfo)

{
    int   CPhase;

    if (!ParentClass::Decode(aStream, TransferInfo))
        return FALSE;
    switch(TransferInfo->TransferMode)
    {
    case TM_ConstData:
        if (!(aStream->Decode(Flags) AND
              aStream->Decode(NrOfPhases)))
            return FALSE;
        if (!Allocate())
            return FALSE;
        
        // The following part is the same for both transfer modes:

    case TM_VarData:
        for (CPhase = 0; CPhase < NrOfPhases; CPhase++)
            if (!Phases[CPhase]->Decode(aStream, TransferInfo))
                return FALSE;
        return TRUE;

    default:
        Warning("invalid TransferMode %d", TransferInfo->TransferMode);
    }
    return TRUE;
}

// ===================================================================================

bool
TIntersectionControl::Encode(TBaseStream *aStream, TTransferInfo *TransferInfo) const

{
    int   CPhase;

    if (!ParentClass::Encode(aStream, TransferInfo))
        return FALSE;
    switch(TransferInfo->TransferMode)
    {
    case TM_ConstData:

        if (!(aStream->Encode(Flags) AND
              aStream->Encode(NrOfPhases)))
            return FALSE;
        
        // The following part is the same for both transfer modes:

    case TM_VarData:
        for (CPhase = 0; CPhase < NrOfPhases; CPhase++)
            if (!Phases[CPhase]->Encode(aStream, TransferInfo))
                return FALSE;
        return TRUE;

    default:
        Warning("invalid TransferMode %d", TransferInfo->TransferMode);
    }
    return TRUE;
}

// ===================================================================================

bool
TIntersectionControl::ExecuteSubTimeStep(const TTimeStep &TimeStep)

{
    int CPhase;

    for (CPhase = 0; CPhase < NrOfPhases; CPhase++)
        if (Phases[CPhase])
            if (!Phases[CPhase]->ExecuteSubTimeStep(TimeStep))
                return FALSE;
    return TRUE;
}

// ===================================================================================

TString
TIntersectionControl::GetClassName(void) const

{
    return "TIntersectionControl";
}

// ===================================================================================

TString
TIntersectionControl::GetInstanceName(void) const

{
    char Line[100];

    sprintf(Line, "%s %d phases", (char *) ParentClass::GetInstanceName(), NrOfPhases);
    return Line;
}

// ===================================================================================

void
TIntersectionControl::_Init(void)

{
    Flags = 0;
    NrOfPhases = 0;
    Phases = 0;
    Random = 0;
}

// ===================================================================================

bool
TIntersectionControl::LoadPhases(TAsciiDataTypeReader *PhaseReader)

{
    int  CPhase;
    long Count;

    if (!PhaseReader->Select(GetParallelID(), Count))
    {
        Error("CreateIntersectionControl: cannot select");
        return FALSE;
    }
    NrOfPhases = Count;
    if (NrOfPhases > 0)
    {
        //      Print("load %ld phases", NrOfPhases);
        if (!Allocate())
            return FALSE;
        for (CPhase = 0; CPhase < NrOfPhases; CPhase++) 
        {
            if ((Phases[CPhase] = (TPhase *) PhaseReader->GetSelected(CPhase)) == 0)
            {
                Error("cannot load phase");
                return FALSE;
            }
            if (Phases[CPhase]->Const.GreenPhaseLength < MIN_PHASE_LENGTH)
                Warning("green phase %d is shorter than minimum", Phases[CPhase]->Const.GreenPhaseLength, MIN_PHASE_LENGTH);
            if (Phases[CPhase]->Const.GreenPhaseLength > MAX_PHASE_LENGTH)
                Warning("green phase %d is longer than maximum", Phases[CPhase]->Const.GreenPhaseLength, MAX_PHASE_LENGTH);
            if (Phases[CPhase]->Const.NonGreenPhaseLength < MIN_PHASE_LENGTH)
                Warning("non-green phase %d is shorter than minimum", Phases[CPhase]->Const.NonGreenPhaseLength, MIN_PHASE_LENGTH);
            if (Phases[CPhase]->Const.NonGreenPhaseLength > MAX_PHASE_LENGTH)
                Warning("non-green phase %d is longer than maximum", Phases[CPhase]->Const.NonGreenPhaseLength, MAX_PHASE_LENGTH);
        }
    }
    return TRUE;
}

// ===================================================================================

bool
TIntersectionControl::SortPhases(void)

{
    int CEdge;

    if (NrOfPhases > 0)
    {
        for (CEdge = 0; CEdge < IncidenceInfo->GetNrOfIncomingEdges(); CEdge++)
        {
            int    CPhase;
            bool   Found;
            TEdge *Edge;
            
            Edge = IncidenceInfo->GetIncomingEdge(CEdge);
            CPhase = 0;
            Found = FALSE;
            while (CPhase < NrOfPhases AND !Found)
                if (Phases[CPhase]->Const.EdgeID == PaminaEdge2TransimsLink(Edge->GetParallelID()))
                    Found = TRUE;
                else
                    CPhase++;
            if (Found)
            {
                if (CEdge != CPhase)
                {
                    TPhase *Temp;
                    
                    Temp = Phases[CEdge];
                    Phases[CEdge] = Phases[CPhase];
                    Phases[CPhase] = Temp;
                }
            }
            else
                Warning("SortPhases: cannot find phase for %s", (char *) Edge->GetObjectName());
        }
    }
    return TRUE;
}

// ===================================================================================
#undef ParentClass
