/*      file     : TrafficData/TMultilaneBoundary.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TMultilaneBoundary.C,v 1.9 2005/03/25 12:28:51 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <TBaseStream.h>
#include <TTransferInfo.h>
#include <TObjectManager.h>
#include <TMultilaneGrid.h>
#include <TVehicle.h>

#include "TMultilaneBoundary.h"

// ***GLOBAL VARIABLES****************************************************************  

TClassInfo *TMultilaneBoundary::ClassInfo = 0;

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TBoundary
// =================================================================================== 

TMultilaneBoundary::TMultilaneBoundary(TMultilaneGrid *Grid, 
                                       const TTimeStep &theTimeStep) :

ParentClass(ClassInfo, Grid)

{        
    _Init();
    TimeStep = theTimeStep;
    NrOfLanes = Grid->GetNrOfLanes();
    BoundaryWidth = Grid->GetBndWidth();
}

// =================================================================================== 

TMultilaneBoundary::TMultilaneBoundary(TParallelID theParallelID) :

ParentClass(ClassInfo, theParallelID)

{    
    _Init();
}

// =================================================================================== 

TMultilaneBoundary::~TMultilaneBoundary(void)

{
    if (Vehicles)
    {   
        Vehicles->DeleteAll(TRUE); // <--- old

        //      TVehicle::UnusedVehicles.Absorb(*Vehicles); // <--- new

        DeleteObject(Vehicles);
    }
}

// =================================================================================== 

bool
TMultilaneBoundary::Assign(void)

{
    long            CLane, CSite;
    TMultilaneGrid *Grid;

    if ((Grid = (TMultilaneGrid *) GetDistributedObject()) == 0)
    {
        Error("Assign: cannot retrieve grid");
        return FALSE;
    }
    for (CLane = 0; CLane < NrOfLanes; CLane++)
        for (CSite = 0; CSite < BoundaryWidth; CSite++)         
            Grid->IncomingBoundary(CLane, CSite) = Sites[CLane][CSite];
    Grid->LoadBoundaryBuffer();
    Grid->ResetBoundaryBuffer();
    if (Vehicles)
    {
        TVehicleEnum CVehicle;

        CVehicle.WalksForwardThrough(Vehicles);
        while (CVehicle())
        {
            //Print(DM_MIGRATION,"Assign: migrated vehicle %s in %ld:%ld", 
            //    (char *) CVehicle->GetInstanceName(),
            //  CVehicle->GetLane(), CVehicle->GetOffset());
            
            if (!Grid->SetMigratedVehicle(CVehicle->GetLane(), CVehicle->GetOffset(),
                                          CVehicle, CVehicle->GetSitePattern()))
            {
                Error("Assign: cannot set vehicle");
                return FALSE;
            }
            CVehicle++;
        }
        Vehicles->DeleteAll();
    }
    return TRUE;
}

// =================================================================================== 

bool
TMultilaneBoundary::Decode(TBaseStream *Stream, TTransferInfo *TransferInfo)

{
    long          CLane;
    TTransferInfo VehicleTransferInfo;

    if (!(ParentClass::Decode(Stream, TransferInfo) AND
          Stream->Decode(NrOfLanes, TMULTILANEBOUNDARY_LONG)))
        return FALSE;
    switch(TransferInfo->TransferMode)
    {
    case TM_SimpleBoundaries:
        for (CLane = 0; CLane < NrOfLanes; CLane++)
            if (!Stream->Decode(Sites[CLane][0], BoundaryWidth))
                return FALSE;
        Vehicles = new TVehicleList;
        VehicleTransferInfo.TransferMode = TM_ConstData;  // !!!!! should be TM_AllData !!!!!
        if (!ObjectManager->ReceiveObjectList((TParallelObjectList *) Vehicles,
                                              Stream, TRUE, &VehicleTransferInfo))
        {
            Error("cannot receive vehicles");
            return FALSE;
        }
        return TRUE;
        
    default:
        debug(Print(DM_CODE, "Decode: TransferMode?"));
    }
    return TRUE;
}

// =================================================================================== 

bool
TMultilaneBoundary::Encode(TBaseStream *Stream, TTransferInfo *TransferInfo) const

{
    long            CLane, CSite;
    TMultilaneGrid *Grid = (TMultilaneGrid *) GetDistributedObject();
    TVehicle       *Vehicle;
    TVehicleList    MigratedVehicles;
    TTransferInfo   VehicleTransferInfo;

    if (!(ParentClass::Encode(Stream, TransferInfo) AND
          Stream->Encode(NrOfLanes, TMULTILANEBOUNDARY_LONG)))
        return FALSE;
    switch (TransferInfo->TransferMode)
    {
    case TM_SimpleBoundaries:
        
        for (CLane = 0; CLane < NrOfLanes; CLane++)
            if (!Stream->Encode(Grid->OutgoingBoundary(CLane, 0), BoundaryWidth))
                return FALSE;
        if (Grid->HasMigratingVehicles(TimeStep))
            for (CLane = 0; CLane < NrOfLanes; CLane++)
                for (CSite = 0; CSite < BoundaryWidth; CSite++)
                    if ((Vehicle = Grid->OutgoingVehicleMigrationArea(CLane, CSite)))
                    {                   
                        Grid->SetBoundaryBuffer(CLane, CSite, 
                                                Grid->OutgoingMigrationArea(CLane, CSite));
                        Vehicle->SetSitePattern(Grid->OutgoingMigrationArea(CLane, CSite));
                        Vehicle->SetLane(CLane);
                        Vehicle->SetOffset(CSite);
                        //Print(DM_MIGRATION, "Encode: migrating vehicle %s on %ld:%ld", 
                        //      (char *) Vehicle->GetInstanceName(), CLane, CSite);
                        Grid->RemoveVehicleFromOutgoingMigrationArea(CLane, CSite);
                        MigratedVehicles.Append(Vehicle);
                    }

        VehicleTransferInfo.TransferMode = TM_ConstData; // !!!!! should be TM_AllData !!!!!

        if (!ObjectManager->SendObjectList((TParallelObjectList *) &MigratedVehicles, 
                                           Stream, &VehicleTransferInfo))
        {
            Error("Encode: cannot send vehicles");
            return FALSE;
        }
        MigratedVehicles.DeleteAll(TRUE);
        return TRUE;
        
    default:
        debug(Print(DM_CODE, "Encode: TransferMode?"));
    }
    return TRUE;
}

// =================================================================================== 

TString
TMultilaneBoundary::GetClassName(void) const

{
    return "TMultilaneBoundary";
}

// =================================================================================== 

TString
TMultilaneBoundary::GetInstanceName(void) const

{
    char Line[256];

    if (Vehicles)
        sprintf(Line, "(%s) Lns:%ld BndW:%ld Veh:%ld", 
                (char *) ParentClass::GetInstanceName(), NrOfLanes, BoundaryWidth,
                Vehicles->GetNrOfElements());
    else
        sprintf(Line, "(%s) Lns:%ld BndW:%ld", (char *) ParentClass::GetInstanceName(),
                NrOfLanes, BoundaryWidth);
    return Line;
}

// =================================================================================== 

void
TMultilaneBoundary::_Init(void)

{
    NrOfLanes = INVALID;
    BoundaryWidth = INVALID;
    Vehicles = 0;
}

// =================================================================================== 
#undef ParentClass
