/*      file     : TrafficData/TRouteplan.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TRouteplan.C,v 1.9 2005/03/25 12:28:52 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <TParser.h>
#include <TBaseStream.h>
#include <TTransferInfo.h>
#include <TRouteplan.h>

// ***GLOBAL VARIABLES****************************************************************

TClassInfo *TRouteplan::ClassInfo = 0;

// ***LOCAL VARIABLES***************************************************************** 

char *RouteplanTokens[] =

{
    "ROUTE"
};

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#  include <TRouteplan.i.C>
#endif

// =================================================================================== 
#define ParentClass TParallelListObject
// ===================================================================================

TRouteplan::TRouteplan(TParallelID theParallelID) : 

    ParentClass(ClassInfo, theParallelID)

{
    _Init();
}

// ===================================================================================

TRouteplan::TRouteplan(long theDepartureTimeStep,  short int theFlags,
                       TParallelID theSourceID, TParallelID theDestID, 
                       TParallelID theRouteID) :

    ParentClass(ClassInfo, INVALID)
    
{
    _Init();
    DepartureTimeStep = theDepartureTimeStep;
    Flags = theFlags;
    SourceID = theSourceID;
    DestID = theDestID;
    RouteID = theRouteID;
}
                       
// ===================================================================================

TRouteplan::TRouteplan(long theDepartureTimeStep, short int theFlags,
                       TParallelID theSourceID, TBaseStream *Stream, 
                       TTransferInfo *TransferInfo) :
    
    ParentClass(ClassInfo, Stream, TransferInfo)
    
{
    _Init();
    DepartureTimeStep = theDepartureTimeStep;
    Flags = theFlags;
    SourceID = theSourceID;
    if (!_Decode(Stream, TransferInfo))
        Fail("cannot decode");
}
                       
// ===================================================================================

TRouteplan::TRouteplan(TBaseStream *Stream, TTransferInfo *TransferInfo) :
    
    ParentClass(ClassInfo, Stream, TransferInfo)
    
{
    _Init();
    if (!_Decode(Stream, TransferInfo))
        Fail("cannot decode");
}
                       
// ===================================================================================

TRouteplan::~TRouteplan(void)

{
}

// ===================================================================================

bool
TRouteplan::AddStep(TParallelID ID, long Estimate)

{
    TRouteStep Step;

    Step.ID = ID;
    Step.Estimate = Estimate;
    if (!Steps.Add(Step))
    {
        Error("AddStep: cannot add");
        return FALSE;
    }
    return TRUE;
}

// ===================================================================================

bool
TRouteplan::ContinuesWith(TRouteStep *RouteSteps, int NrOfRouteSteps) const

{
    int CSubStep, CMyStep;

    CSubStep = 0;
    CMyStep = GetCurrentStep();
    while (CMyStep < GetNrOfWholeSteps() AND CSubStep < NrOfRouteSteps)
    {
        if (RouteSteps[CSubStep].ID != Steps[CMyStep].ID)
            return FALSE;
        CSubStep++;
        CMyStep++;
    }
    return CSubStep == NrOfRouteSteps;
}

// ===================================================================================

TBaseObject *
TRouteplan::Copy(void) const

{
    TRouteplan *NewRouteplan;

    if ((NewRouteplan = new TRouteplan(GetParallelID())))
    {
        NewRouteplan->Flags = Flags;
        NewRouteplan->CurrentStep = CurrentStep;
        NewRouteplan->DepartureTimeStep = DepartureTimeStep;
        NewRouteplan->InsertionTimeStep = InsertionTimeStep;
        NewRouteplan->DeletionTimeStep = DeletionTimeStep;
        for (int i = 0; i < Steps.GetNrOfItems(); i++)
            NewRouteplan->Steps.Add(Steps[i]);
    }
    return NewRouteplan;
}

// ===================================================================================

bool
TRouteplan::_Decode(TBaseStream *aStream, TTransferInfo *TransferInfo)

{
    switch(TransferInfo? TransferInfo->TransferMode : TM_VarData)
    {
    case TM_VarData:
    case TM_ConstData:

        if (!(aStream->Decode(DepartureTimeStep, TROUTEPLAN_CONST_LONG) AND
              aStream->Decode(SourceID, TROUTEPLAN_CONST_LONG_ID) AND
              aStream->Decode(CurrentStep, TROUTEPLAN_VAR_LONG) AND
              aStream->Decode(EstimatedRelativeDelay, TROUTEPLAN_VAR_FLOAT) AND
              DecodeSteps(aStream)))
        {
            Error("cannot decode var data");
            return FALSE;
        }
        break;
        
    case TM_AsciiData:

        if (!(aStream->Decode(DestID) AND
              aStream->Decode(RouteID) AND
              DecodeSteps(aStream)))
        {
            Error("cannot decode const data");
            return FALSE;
        }
        break;

    default:

        debug(Print(DM_CODE, "Decode: TransferMode?"));
    }
    return TRUE;
}

// ===================================================================================

bool
TRouteplan::Decode(TBaseStream *aStream, TTransferInfo *TransferInfo)

{
    return (ParentClass::Decode(aStream, TransferInfo) AND
            _Decode(aStream, TransferInfo));
}

// ===================================================================================

bool
TRouteplan::DecodeSteps(TBaseStream *Stream)

{
    int NrOfSteps;
    
    if (!Stream->Decode(NrOfSteps))
    {
        Error("cannot read number of plan steps");
        return FALSE;
    }
    if (NrOfSteps > MAX_ROUTE_LENGTH)
    {
        Error("route length %ld excedes maximum %ld", NrOfSteps, MAX_ROUTE_LENGTH);
        return FALSE;
    }
    Steps.Resize(NrOfSteps);
    for (long i = 0; i < NrOfSteps; i++)
        if (!(Stream->Decode(Steps[i].ID) &&
              Stream->Decode(Steps[i].Estimate)))
        {           
            Error("cannot read plan steps");
            return FALSE;
        }
    return TRUE;
}

// ===================================================================================

bool
TRouteplan::Detour(long CurrentTimeStep, long LastValidStep, TRouteStep *DetourSteps, 
                   long NrOfDetourSteps, long DetourArrivalTime)

{
    long OldLength;
    long NewLength;
    long NrOfReplacedSteps;
    long CStep;
    
    if (LastValidStep >= GetNrOfSteps() - 1)
    {
        Error("Detour: LastValidStep >= NrOfSteps - 1");
        return FALSE;
    }
    if (LastValidStep < CurrentStep)
    {
        Error("Detour: LastValidStep <= CurrentStep");
        return FALSE;
    }
    OldLength = GetNrOfSteps();
    NrOfReplacedSteps = LastValidStep - CurrentStep + 1;
    NewLength = GetNrOfSteps() + NrOfDetourSteps - NrOfReplacedSteps;
    if (NrOfReplacedSteps != NrOfDetourSteps)
    {
        if (NewLength > OldLength)
            if (!Steps.Resize(NewLength))
            {
                Error("Reroute: cannot resize plan to new length %ld", NewLength);
                return FALSE;
            }
        memmove((void *) &Steps[CurrentStep + NrOfDetourSteps].ID, 
                (void *) &Steps[LastValidStep + 1].ID,
                (OldLength - LastValidStep - 1) * sizeof(TRouteStep));
        if (NewLength < OldLength)
            if (!Steps.Resize(NewLength))
            {
                Error("Reroute: cannot resize plan to new length %ld", NewLength);
                return FALSE;
            }       
    }
    memcpy((void *) &Steps[CurrentStep].ID,
           (void *) &DetourSteps[0].ID, NrOfDetourSteps * sizeof(TRouteStep));
    for (CStep = CurrentStep + NrOfDetourSteps + 1; CStep < NewLength; CStep++)
    {
        long TravelTime;

        TravelTime = Steps[CStep].Estimate - Steps[CStep - 1].Estimate;
        Steps[CStep - 1].Estimate = DetourArrivalTime;
        DetourArrivalTime += TravelTime;
    }
    NrOfReroutes++;
    LatestRerouteTimeStep = CurrentTimeStep;
    if (OriginalEstimatedTravelTime == INVALID)
        OriginalEstimatedTravelTime = EstimatedTravelTime;
    EstimatedTravelTime = DetourArrivalTime - DepartureTimeStep;
    EstimatedRelativeDelay = 0.0; // ?????
    return TRUE;
}

// ===================================================================================

bool
TRouteplan::Encode(TBaseStream *aStream, TTransferInfo *TransferInfo) const

{
    long EncodeCurrentStep;
    long StepOffset = 0;

    if (!ParentClass::Encode(aStream, TransferInfo))
        return FALSE;
    switch(TransferInfo? TransferInfo->TransferMode : TM_VarData)
    {
    case TM_VarData:
    case TM_ConstData:

        if (Flags & RP_FLAG_SHORTEN_ON_ENCODE)
        {
            EncodeCurrentStep = 0;
            StepOffset = CurrentStep;
        }
        else
        {
            EncodeCurrentStep = CurrentStep;
            StepOffset = 0;
        }
        if (!(aStream->Encode(DepartureTimeStep, TROUTEPLAN_CONST_LONG) AND
              aStream->Encode(SourceID, TROUTEPLAN_CONST_LONG_ID) AND
              aStream->Encode(EncodeCurrentStep) AND
              aStream->Encode(InsertionTimeStep,  
                              TROUTEPLAN_VAR_LONG - 1) AND
              aStream->Encode(EstimatedRelativeDelay, TROUTEPLAN_VAR_FLOAT) AND
              EncodeSteps(aStream, StepOffset)))
        {
            Error("cannot encode const/var data");
                return FALSE;
        }
        break;
        
    case TM_AsciiData:
                
        if (!aStream->IsBinary())
            aStream->Write("\n%s\n\t", RouteplanTokens[TOKEN_ROUTE]);
        if (!(aStream->Encode(DepartureTimeStep, TROUTEPLAN_CONST_LONG) AND
              aStream->Encode(SourceID, TROUTEPLAN_CONST_LONG_ID) AND
              EncodeSteps(aStream, StepOffset)))
        {
            Error("cannot encode ASCII data");
            return FALSE;
        }
        break;

    default:

        debug(Print(DM_CODE, "Encode: TransferMode?"));
    }
    return TRUE;
}

// ===================================================================================

bool
TRouteplan::EncodeSteps(TBaseStream *aStream, int StepOffset) const

{
    int NrOfSteps;

    NrOfSteps = Steps.GetNrOfItems() - StepOffset;
    if (!(aStream->Encode(NrOfSteps)))
        return FALSE;
    for (int i = 0; i < NrOfSteps; i++)
        if (!(aStream->Encode(Steps[StepOffset + i].ID) &&
              aStream->Encode(Steps[StepOffset + i].Estimate)))
            return FALSE;
    return TRUE;
}

// ===================================================================================

TString
TRouteplan::GetClassName(void) const

{
    return "TRouteplan";
}

// ===================================================================================

TString
TRouteplan::GetInstanceName(void) const

{
    char        Line[128];
    static char Format[64] = { 0 };

    if (0 == Format[0]) {
        sprintf(Format, "F:%%ld Dprt:%%ld Src:%s Dst:%s %%ld/%%ld CID:%s", 
                PARALLEL_ID_PRINTF_FORMAT, PARALLEL_ID_PRINTF_FORMAT, PARALLEL_ID_PRINTF_FORMAT);
    }
    sprintf(Line, Format, Flags, GetDepartureTimeStep(),
            SourceID, DestID, CurrentStep + 1, GetPlanLength(), GetCurrentID());
    return Line;
}

// =================================================================================== 

TParallelID
TRouteplan::GetCurrentEdgeID(void)

{
    if (CurrentStep == 0)
    {
        Error("GetCurrentEdgeID: not yet entered first edge?");
        return INVALID;
    }
    return Steps[CurrentStep - 1].ID;
}

// =================================================================================== 

void
TRouteplan::_Init(void)

{
    Flags = 0;
    CurrentStep = 0;
    DepartureTimeStep = INVALID;
    InsertionTimeStep = INVALID;
    DeletionTimeStep = INVALID;
    EstimatedTravelTime = INVALID;
    OriginalEstimatedTravelTime = INVALID;
    NrOfReroutes = 0;
    LatestRerouteTimeStep = INVALID;
    EstimatedRelativeDelay = (float) INVALID;
}

// ===================================================================================

bool
TRouteplan::Reroute(long CurrentTimeStep, TRouteStep *RouteSteps, int NrOfRouteSteps, 
                    long RerouteArrivalTime)

{
    long       NewLength;
    TRouteStep LastStep;

    NewLength = GetCurrentStep() + NrOfRouteSteps;
    if (DestinationIsAccessory())
    {
        NewLength++;
        LastStep = Steps[Steps.GetNrOfItems() - 1];
    }
    if (!Steps.Resize(NewLength))
    {
        Error("Reroute: cannot resize plan to new length %ld", NewLength);
        return FALSE;
    }
    if (DestinationIsAccessory())
        Steps[NewLength - 1] = LastStep;
    memcpy(& Steps[GetCurrentStep()], RouteSteps, NrOfRouteSteps * sizeof(TRouteStep));
    NrOfReroutes++;
    LatestRerouteTimeStep = CurrentTimeStep;
    if (OriginalEstimatedTravelTime == INVALID)
        OriginalEstimatedTravelTime = EstimatedTravelTime;
    EstimatedTravelTime = RerouteArrivalTime - DepartureTimeStep;
    EstimatedRelativeDelay = 0.0; // ?????
    return TRUE;
}

// ===================================================================================

bool
TRouteplan::Skip(TBaseStream *Stream)

{
    long       NrOfSteps;
    TRouteStep DummySteps[MAX_ROUTE_LENGTH];
    
    if (!(Stream->Decode(DestID, TROUTEPLAN_CONST_LONG_ID - 1) AND
          Stream->Decode(NrOfSteps)))
    {
        Error("Skip: cannot read plan header");
        return FALSE;
    }
    if (NrOfSteps > MAX_ROUTE_LENGTH)
    {
        Error("Skip: route length %ld excedes maximum %ld", NrOfSteps, MAX_ROUTE_LENGTH);
        return FALSE;
    }
    for (int i = 0; i < NrOfSteps; i++)
        if (!(Stream->Decode(DummySteps[i].ID) &&
              Stream->Decode(DummySteps[i].Estimate)))
        {           
            Error("Skip: cannot read plan steps");
            return FALSE;
        }
    return TRUE;
}

// ===================================================================================

void
TRouteplan::Step(void)

{
    if (CurrentStep < GetPlanLength())
        CurrentStep++;
}

// ===================================================================================

bool
TRouteplan::TerminatesWith(TRouteStep *RouteSteps, int NrOfRouteSteps) const

{
    int CMyStep, CSubStep;

    CMyStep = GetCurrentStep();
    CSubStep = 0;
    while (CMyStep < GetNrOfWholeSteps() AND CSubStep < NrOfRouteSteps)
    {
        if (RouteSteps[CSubStep].ID != Steps[CMyStep].ID)
            return FALSE;
        CSubStep++;
        CMyStep++;
    }
    return (CMyStep == GetNrOfWholeSteps() AND CSubStep == NrOfRouteSteps);
}

// =================================================================================== 
/*
bool
TRouteplan::Write(TBaseStream *Stream)

{
    long i;
    
    Stream->Write("\n%s\t%ld %d %ld\n\t%ld %ld %ld\n\t", RouteplanTokens[TOKEN_ROUTE],
                  DepartureTimeStep, Flags, SourceID, 
                  DestID, RouteID, Steps.GetNrOfItems());
    for (i = 0; i < Steps.GetNrOfItems(); i++)
    {
        Stream->Write("%ld %ld ", Steps[i].ID, Steps[i].Estimate);
        if (i % 4 == 3)
            Stream->Write("\n\t");
    }
    if (i % 4 != 0)
        Stream->Write("\n");
    return TRUE;
}
*/
// =================================================================================== 

bool
TRouteplan::WriteStat(TBaseStream *Stream)

{
    long ActualTravelTime;
    long InsertionDelay, TravelDelay;

    // the output is:
    //
    // scheduled departure time step
    // estimated travel time
    // insertion time step
    // time spent in insertion queue
    // arrival time step
    // actual travel time 
    // absolute deviation from estimated travel time (departure time)
    // relative deviation from estimated travel time (departure time)
    
    ActualTravelTime = DeletionTimeStep - InsertionTimeStep;
    InsertionDelay =  InsertionTimeStep - DepartureTimeStep;
    TravelDelay = ActualTravelTime - EstimatedTravelTime;
    if (EstimatedTravelTime > 0)
        Stream->Write("%ld %ld %ld %ld %ld %ld %ld %.3f %ld %ld %ld\n",
                      DepartureTimeStep,
                      EstimatedTravelTime,
                      InsertionTimeStep,
                      InsertionDelay,
                      DeletionTimeStep,
                      ActualTravelTime,
                      TravelDelay,
                      1.0 * TravelDelay / EstimatedTravelTime,
                      RouteID,
                      NrOfReroutes,
                      OriginalEstimatedTravelTime);
    else
        Stream->Write("# vehicle with start time %ld has "
                      "zero estimated travel time!\n", DepartureTimeStep);
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
