/*      file     : common/TDLList.h
        project  : General Toolbox
        author   : Marcus Rickert
        objects  : TListNodePrimitive
                   TDLNodePrimitive
                   TSLNodePrimitive
                   TDLListPrimitive
                   TDLList<>
                   TDLListReference
                   TDLListEnumPrimitive
                   TDLListEnum<>
        RCS      : $Id: TDLList.h,v 1.16 2005/03/25 11:04:59 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

#include "gcc_pragma_interface.h"

#ifndef TDLLIST_H
#define TDLLIST_H

/* ***INCLUDES************************************************************************ */

#include <Basic.h>
#include <TContainer.h>
#include <TArray.h>

// ***FORWARDS************************************************************************

class TDLListPrimitive;
class TListNodePrimitive;
template < class T > class TDLList;

// ***CLASSES*************************************************************************

//extern "C"
struct TDLListReference

{
    short int                 PrevIndex;
    short int                 NextIndex;
    const TListNodePrimitive *Prev;
    const TListNodePrimitive *Next;
    TDLListPrimitive         *List;
};

/* =================================================================================== */

/** Base class for classes that are to be inserted into intrusive lists. 
*/

class TListNodePrimitive

{
 public:

    virtual void
        FreeReference(int Index) const = 0;
    inline bool
        IsInList(const TDLListPrimitive *aList) const;
    bool
        LeftLink(TDLListPrimitive *theList, const TListNodePrimitive *aNode, 
                 bool Unique = FALSE) const;
    bool
        Unlink(TDLListPrimitive *aList) const;
    virtual bool
        UnlinkAll(void) = 0;
    virtual bool
        GetFreeReferenceIndex(int &Index) const = 0;
    virtual bool
        GetListIndex(const TDLListPrimitive *theList, int &Index) const = 0;
    virtual const TListNodePrimitive *
        GetNext(int Index) const = 0;
    virtual int
        GetNextIndex(int Index) const = 0;
    virtual const TListNodePrimitive *
        GetPrev(int Index) const = 0;
    virtual int
        GetPrevIndex(int Index) const = 0;
    virtual void
        SetList(int Index, TDLListPrimitive *theList) const = 0;
    virtual void
        SetNext(int Index, const TListNodePrimitive *NextNode, int NextIndex) const = 0;
    virtual void
        SetPrev(int Index, const TListNodePrimitive *PrevNode, int PrevIndex) const = 0;
};

/* =================================================================================== */

/** Base class for classes that are to be inserted into singly-linked intrusive lists. 
*/

class TSLNodePrimitive : public TListNodePrimitive

{
public:

    TSLNodePrimitive(void);
    virtual ~TSLNodePrimitive(void);
    
    virtual void
        FreeReference(int Index) const;
    virtual bool
        GetFreeReferenceIndex(int &Index) const;
    virtual bool
        GetListIndex(const TDLListPrimitive *theList, int &Index) const;
    virtual const TListNodePrimitive *
        GetNext(int Index) const;
    virtual int
        GetNextIndex(int Index) const;
    virtual const TListNodePrimitive *
        GetPrev(int Index) const;
    virtual int
        GetPrevIndex(int Index) const;
    virtual void
        SetList(int Index, TDLListPrimitive *theList) const;
    virtual void
        SetNext(int Index, const TListNodePrimitive *NextNode, int NextIndex) const;
    virtual void
        SetPrev(int Index, const TListNodePrimitive *PrevNode, int PrevIndex) const;
    virtual bool
        UnlinkAll(void);

private:
    
    MUTABLE_DECL TDLListReference Ref;
};

/* =================================================================================== */

/** Base class for classes that are to be inserted into intrusive doubly-linked lists.
*/

class TDLNodePrimitive : public TListNodePrimitive

{
 public:

    TDLNodePrimitive(void);
    virtual ~TDLNodePrimitive(void);

    virtual void
        FreeReference(int Index) const;
    virtual bool
        GetFreeReferenceIndex(int &Index) const;
    virtual bool
        GetListIndex(const TDLListPrimitive *theList, int &Index) const;
    virtual const TListNodePrimitive *
        GetNext(int Index) const;
    virtual int
        GetNextIndex(int Index) const;
    virtual const TListNodePrimitive *
        GetPrev(int Index) const;
    virtual int
        GetPrevIndex(int Index) const;
    virtual void
        SetList(int Index, TDLListPrimitive *theList) const;
    virtual void
        SetNext(int Index, const TListNodePrimitive *NextNode, int NextIndex) const;
    virtual void
        SetPrev(int Index, const TListNodePrimitive *PrevNode, int PrevIndex) const;
    virtual bool
        UnlinkAll(void);

 private:

    MUTABLE_DECL short int                FirstFreeIndex;
    MUTABLE_DECL TArray<TDLListReference> Refs;
};

/* =================================================================================== */

class TDLListEnumPrimitive

{
 public:

    TDLListEnumPrimitive(void);
    virtual ~TDLListEnumPrimitive(void);

    operator long () const;
    operator TListNodePrimitive const & () const;
    operator TListNodePrimitive & () const;
    operator TListNodePrimitive const * () const;
    operator TListNodePrimitive * () const;
    
    void
        operator = (TListNodePrimitive *Node);

    TListNodePrimitive *
        operator -> ();
    inline bool
        operator () (void);
    inline void
        operator ++(int);
    inline void
        operator --(int);

    inline int
    GetCurrentIndex(void) const;

    void
        UnlinkList(void);
    void
        UpdateUponDelete(const TListNodePrimitive *Previous, int PreviousIndex);
    void
        _WalksBackwardThrough(TDLListPrimitive &theList);
    void
        _WalksForwardThrough(TDLListPrimitive &theList);

 protected:
    
    const TListNodePrimitive *Current;

    void
        Step(void);

private:

    const TListNodePrimitive *Head;
    int                       NextIndex;
    int                       CurrentIndex;
    const TListNodePrimitive *Next;
    TDLListPrimitive         *CurrentList;
    TDLListEnumPrimitive     *PreviousEnum;
};

/* =================================================================================== */

class TDLListPrimitive : public TContainer

{
 public:

    void
    _Absorb(TDLListPrimitive &OtherList);

    bool
        _Append(TDLListPrimitive &aList, bool Unique = FALSE);
    bool
        DeleteAll(bool DisposeElements = FALSE);
    inline TDLListEnumPrimitive *
        GetCurrentEnum(void);
    inline long
        GetNrOfElements(void) const;
    inline void
        UpdateCurrentEnum(const TListNodePrimitive *Previous, int PreviousIndex);
    inline void
        SetCurrentEnum(TDLListEnumPrimitive *theEnum);

 protected:

    TDLListPrimitive(void);
    virtual ~TDLListPrimitive(void);

    bool
        _Append(const TListNodePrimitive *anElement, bool Unique = FALSE);
    bool
        _Delete(const TListNodePrimitive *anElement);
    inline TListNodePrimitive *
        GetHead(void) const;
    virtual TString
        GetInstanceName(void) const;
    bool
        _IsInList(const TListNodePrimitive *anElement) const;
    bool
        _Prepend(const TListNodePrimitive *anElement, bool Unique = FALSE);

 private:

    friend class TDLListEnumPrimitive;
    friend class TListNodePrimitive;
    
    TSLNodePrimitive    Head;

    long                  NrOfElements;
    TDLListEnumPrimitive *CurrentEnum;    
};

/* =================================================================================== */

template <class T>
class TDLListEnum : public TDLListEnumPrimitive

{
 public:

    virtual ~TDLListEnum(void);

    operator T & ();
    operator T * ();
    operator const T * () const;
    T *
        operator -> ();
    void
        operator = (T *Node);
    void
        WalksBackwardThrough(const TDLList<T> &theList);
    void
        WalksBackwardThrough(const TDLList<T> *theList);
    void
        WalksForwardThrough(const TDLList<T> &theList);
    void
        WalksForwardThrough(const TDLList<T> *theList);
};

/* =================================================================================== */

template <class T>
class TDLList : public TDLListPrimitive

{
 public:

//    TDLList(void) { };
//    virtual ~TDLList(void) { };

    void
    Absorb(TDLList< T > &aList);

    bool
        Append(TDLList< T > &aList, bool Unique = FALSE);
    bool
        Append(const T *anElement, bool Unique = FALSE);
    bool
        Delete(const T *anElement);
    virtual TString
        GetClassName(void) const;
    T *
        GetFirst(void);
    T *
        GetLast(void);
    bool
        IsInList(const T *anElement) const;
    bool
        Prepend(const T *anElement, bool Unique = FALSE);
};

// ***INLINE METHODS******************************************************************

#ifndef NO_INLINE
#include "TDLList.i.C"
#endif

// ***TEMPLATE METHODS****************************************************************

#if defined(INCLUDE_TEMPLATES) || defined(TEMPLATEBODIES)
#include "TDLList.t.C"
#endif

// ***PROTOTYPES********************************************************************** 
#ifdef __cplusplus
extern "C" {
#endif
    void
        InitListReference(TDLListReference &Ref);
#ifdef __cplusplus
}
#endif
// ===================================================================================
// Local Variables:
// mode:c++
// End:
#endif
