/*      file     : common/TLocation.h
        project  : General Toolbox
        author   : Marcus Rickert
        objects  : TLocation
                   TRectangle
        RCS      : $Id: TLocation.h,v 1.7 2005/03/25 11:04:59 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

#ifndef TLOCATION_H
#define TLOCATION_H

// ***INCLUDES************************************************************************

#include <TBaseObject.h>

// ***DEFINES*************************************************************************

#define MAX_COORDINATE 1.0e99
#define MIN_COORDINATE (-1.0e99)
#define LengthInKilometer(Length) ((Length) / 100000.0)
#define KilometerToLength(Length) ((Length) * 100000.0)
#define LengthInMeter(Length) ((Length) / 100.0)
#define MeterToLength(Length) ((Length) * 100.0)

// ***TYPES***************************************************************************

typedef double TCoordinate;

// ***FORWARDS************************************************************************

class TBaseStream;

// ***CLASSES*************************************************************************

/** Class holding a geometric location.
  */

class TLocation : public TBaseObject

{
 private:

    /** X coordinate given in Centimeters. */
    TCoordinate _X;

    /** Y coordinate given in Centimeters. */
    TCoordinate _Y;

 public:

    TLocation(const TCoordinate X = 0.0, const TCoordinate Y = 0.0);
    
    void
        operator += (const TLocation &aLocation);
    void
        operator -= (const TLocation &aLocation);

    TLocation
        operator - (const TLocation &aLocation) const;

    TLocation
        operator + (const TLocation &aLocation) const;

    void
    operator *= (double Factor);
        
    void
    operator /= (double Divisor);
        
    double
        AngleTo(TLocation &OtherLocation) const;
    bool
        Decode(TBaseStream &aStream);
    TCoordinate
        DistanceTo(const TLocation &aLocation) const;
    bool
        Encode(TBaseStream &aStream) const;
    virtual TString
        GetClassName(void) const;
    virtual TString
        GetInstanceName(void) const;

    double 
    Length(void) const;

    inline TCoordinate
    ManhattenDistanceTo(const TLocation &OtherLoc) const;

    TLocation
    Normal(void) const;

    void
        Reset(void);
    void
        Set(TCoordinate theX, TCoordinate theY);
    void
        SetX(TCoordinate theX);
    void
        SetY(TCoordinate theY);      
    inline TCoordinate 
        x(void) const;
    inline TCoordinate 
        y(void) const;
    inline TCoordinate &
        x(void);
    inline TCoordinate &
        y(void);
};

/* =================================================================================== */

/** Class holding a geometric rectangle.
 */

class TRectangle : public TBaseObject

{
 private:

    TLocation Locations[2];

 public:

    TRectangle(void) { };
    TRectangle(const TLocation Location1, const TLocation Location2);

    void
    operator *= (double Factor);

    void
    operator += (double Frame);

    void
    operator += (const TLocation &Vector);

    TLocation
    Center(void) const;

    void
        CenterAround(const TLocation &Location);
    bool
        Contains(const TLocation &Location) const;
    bool
        Decode(TBaseStream &aStream);
    bool
        Encode(TBaseStream &aStream) const;
    void
        Empty(void);
    virtual TString
        GetClassName(void) const;
    TCoordinate
        GetDeltaX(void) const;
    TCoordinate
        GetDeltaY(void) const;
    virtual TString
        GetInstanceName(void) const;
    void
        GetLowerLeftLocation(TLocation &Location) const;
    TCoordinate &
        GetMaxX(void); 
    TCoordinate
        GetMaxX(void) const; 
    TCoordinate &
        GetMaxY(void);
    TCoordinate
        GetMaxY(void) const;
    TCoordinate &
        GetMinX(void);
    TCoordinate
        GetMinX(void) const;
    TCoordinate &
        GetMinY(void);
    TCoordinate
        GetMinY(void) const;
    void
        GetUpperRightLocation(TLocation &Location) const;
    void
        Merge(const TLocation &Location);
    void
        Merge(const TRectangle &Rectangle);
    bool
        Overlaps(const TRectangle &OtherRect) const;
};

// ***PROTOTYPES**********************************************************************

TLocation
operator * (double Factor, const TLocation &Loc);

// ***INLINE METHODS******************************************************************

#ifndef NO_INLINE
#  include "TLocation.i.C"
#endif

// =================================================================================== 
// Local Variables:
// mode:c++
// End:
#endif
