/*      file     : mpi/TMPIMessageControl.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TMPIMessageControl.C,v 1.14 2005/05/01 14:57:58 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <stdlib.h>

#include <TEnvironment.h>
#include <TConfiguration.h>

#include <TParallel.h>
#include <TCPNInfo.h>
#include <TDestination.h>

#include "Mpi.h"
#include "MPIConfig.h"
#include "TMPIBuffer.h"
#include "TMPIMessage.h"
#include "TMPIMessageControl.h"

// ***DEFINES************************************************************************* 

#define MPI_MAXBUFFER        8L
#define MPI_BUFFERSIZE       (1024L*1024L)

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TCommunicationControl
// =================================================================================== 

TMPIMessageControl::TMPIMessageControl(void)

{
    _Init();
}

// =================================================================================== 

TMPIMessageControl::~TMPIMessageControl(void)

{
    MPIFreeBuffers.DeleteAll(TRUE);
    MPIUsedBuffers.DeleteAll(TRUE);
}

// =================================================================================== 

bool
TMPIMessageControl::Activate(void)

{    
  TConfiguration *C = Environment->GetConfiguration();

  if (!ParentClass::Activate())
      return FALSE;
  LoadBaseTables();
  MaxBuffer = (int) C->GetLong(MPI_GROUP, MPI_NAME_MAXBUFFER);
  BufferSize = (int) C->GetLong(MPI_GROUP, MPI_NAME_BUFFERSIZE);
  return TRUE;
}

// =================================================================================== 

bool
TMPIMessageControl::GetBuffer(TMPIBuffer *&Buffer)

{
    while (MPIFreeBuffers.GetNrOfElements() > MaxBuffer)
        delete MPIFreeBuffers.GetFirst();
    if (MPIFreeBuffers.GetNrOfElements() > 0)
        Buffer = MPIFreeBuffers.GetFirst();
    else        
        if ((Buffer = new TMPIBuffer(BufferSize)) == 0 OR
            Buffer->Failed())
        {
            Error("GetNewMessage: cannot allocate message buffer");
            return FALSE;
        }
    return TRUE;
}

// =================================================================================== 

long
TMPIMessageControl::GetCapabilities(void) const

{
    return CC_SEND_MESSAGE | CC_LINEAR_BROADCAST;
}

// =================================================================================== 

TString
TMPIMessageControl::GetClassName(void) const

{
    return "TMPIMessageControl";
}

// =================================================================================== 

TBaseMessage *
TMPIMessageControl::GetNewMessage(TMessageID theID, TCPNID theSender,
                                  TDestination *theDestination, TMessageTag theTag, 
                                  TMessageType theType, TMessageFlags theFlags)

{
    TMPIBuffer *Buffer = 0;
    if (!GetBuffer(Buffer))
        return 0;
    return new TMPIMessage(this, Buffer, theID, theSender,theDestination, theTag, 
                           theType, theFlags);
}

// =================================================================================== 

void
TMPIMessageControl::_Init(void)

{
  TConfiguration *C = Environment->GetConfiguration();
  
  C->Set(TConfigEntry(MPI_GROUP, MPI_NAME_MAXBUFFER, MPI_MAXBUFFER));
  C->Set(TConfigEntry(MPI_GROUP, MPI_NAME_BUFFERSIZE, MPI_BUFFERSIZE));
}

// =================================================================================== 

bool
TMPIMessageControl::IsBlockable(void) const

{
    return TRUE;
}

// =================================================================================== 

void
TMPIMessageControl::LoadBaseTables(void)

{
    int i;
    
    for (i = 0; i < BaseTypeCount; i++)
        BaseTypes[i] = 0;
    BaseTypes[BaseTypeShort] = MPI_SHORT;
    BaseTypes[BaseTypeUShort] = MPI_UNSIGNED_SHORT;
    BaseTypes[BaseTypeInt] = MPI_INT;
    BaseTypes[BaseTypeUInt] = MPI_UNSIGNED;
    BaseTypes[BaseTypeLong] = MPI_LONG;
    BaseTypes[BaseTypeULong] = MPI_UNSIGNED_LONG;
    BaseTypes[BaseTypeFloat] = MPI_FLOAT;
    BaseTypes[BaseTypeDouble] = MPI_DOUBLE;

    for (i = 0; i < BaseOperationCount; i++)
        BaseOperations[i] = 0;
    BaseOperations[OpSum] = MPI_SUM;
    BaseOperations[OpProduct] = MPI_PROD;
}

// =================================================================================== 

bool
TMPIMessageControl::MessageIsPending(void) const

{
    int        Flag = FALSE;
    MPI_Status MPIStatus;

    MPIStatus.MPI_SOURCE = INVALID;
    MPIStatus.MPI_TAG = INVALID;
    MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &Flag, &MPIStatus);
    debug(Print(DM_MSG_CONTROL, "MessageIsPending: MPI_Iprobe returns Flag:%d "
                "Source:%d Tag:%d", Flag, MPIStatus.MPI_SOURCE, MPIStatus.MPI_TAG));
    return Flag;
}

// =================================================================================== 

bool
TMPIMessageControl::ReceiveMessage(TBaseMessage *&Message, long)

{
    MPI_Status  MPIStatus;
    TMPIBuffer *Buffer = 0;
    int         Count;

    Message = 0;
    if (!GetBuffer(Buffer))
        return 0;
    MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &MPIStatus);
    debug(Print(DM_MSG_CONTROL, "ReceiveMessage: receiving MPI message Source:%d"
                " Tag:%d", MPIStatus.MPI_SOURCE, MPIStatus.MPI_TAG));
    MPI_Get_count(&MPIStatus, MPI_PACKED, &Count);
    debug(Print(DM_MSG_CONTROL, "ReceiveMessage: MPI_Get_count reports length %d",
                Count));
    if ((Message = new TMPIMessage(this, Buffer, MPIStatus.MPI_SOURCE)) == 0)
    {
        Error("ReceiveMessage: cannot create new message");
        return FALSE;
    }
    MPI_Recv(Buffer->GetBase(), Count, MPI_PACKED, MPI_ANY_SOURCE, MPI_ANY_TAG,
             MPI_COMM_WORLD, &MPIStatus);
    return Message->BeginReceive();
}

// =================================================================================== 

bool
TMPIMessageControl::ReduceLinearArray(const void* SendBuffer, void* ReceiveBuffer, 
                                      int Count, TBaseType BaseType, 
                                      TBaseOperation BaseOperation, TCPNID Root)

{
    if (BaseOperations[BaseOperation] == 0)
    {
        Error("ReduceLinearArray: base operation %d not supported", BaseOperation);
        return FALSE;
    }
    if (BaseTypes[BaseType] == 0)
    {
        Error("ReduceLinearArray: base data type not supported");
        return FALSE;
    }
    MPI_Reduce((void *) SendBuffer, ReceiveBuffer, Count, BaseTypes[BaseType],
               BaseOperations[BaseOperation], (int) Root, MPI_COMM_WORLD);
    return TRUE;
}

/* =================================================================================== */

bool
TMPIMessageControl::SetEncoding(TEncodingMode)

{
    return TRUE;
}

/* =================================================================================== */
/*
void
TMPIMessageControl::WriteSignature(void) const

{
    SignatureHeader(GetClassName());
    SignatureLine("MaxBuffer      = %d", MaxBuffer);
    SignatureLine("BufferSize     = %d [byte]", BufferSize);
}
*/
// =================================================================================== 
#undef ParentClass
