/*      file     : parallel/TCommunication
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TCommunication.C,v 1.24 2005/03/25 11:06:20 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <string.h>

#include <ParallelConfig.h>
#include <TString.h>
#include <TTimer.h>
#include <TEnvironment.h>
#include <TParallel.h>
#include <TConfiguration.h>

#ifdef USE_PVM
#  include <TPVMMessageControl.h>
#elif  USE_MPI
#  include <TMPIMessageControl.h>
#endif
#include "TMemoryMessageControl.h"
#include "TBaseMessage.h"
#include "TMessageRecipient.h"
#include "TCommunication.h"

// ***DEFINES************************************************************************* 

#define DEFAULT_ENCODING EncodingRaw
#define PARALLEL_DEFAULT_MESSAGETIMEOUT 0.0

// ***GLOBAL VARIABLES****************************************************************  

TCommunication *Communication = 0;

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#endif

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

#include "Reduce.Bodies.C"

// =================================================================================== 


TCommunication::TCommunication(void)

{
    if (Environment == 0)
    {
        Fail("no TEnvironment active");
    }
    if (Communication)
    {
        Fail("another instance of TCommunication already active");
    }
    Communication = this;
    _Init();
    debug(Print(DM_LIFESPAN, "CONSTRUCTOR"));
    Configuration->SetDefault(TConfigEntry(PARALLEL_GROUP, PARALLEL_NAME_MESSAGETIMEOUT,
                                           PARALLEL_DEFAULT_MESSAGETIMEOUT));
}

// =================================================================================== 

TCommunication::~TCommunication(void)

{
    debug(Print(DM_LIFESPAN, "DESTRUCTOR"));
    Deactivate();
}

// =================================================================================== 

bool
TCommunication::Activate(void)

{
    int i;

    debug(Print(DM_LIFESPAN, "Activate"));
    for (i = 0; i < NrOfControllers; i++)
    {
        if (Controllers[i] == 0)
        {
            Error("could not instantiate controller #%d", i);
            return FALSE;
        }
        if (!Controllers[i]->Activate())
        {
            Error("cannot activate controller %s", 
                  (char *) Controllers[i]->GetClassName());
            return FALSE;
        }
        Controllers[i]->SetEncoding(DEFAULT_ENCODING);
    }
    MessageTimeOut = Configuration->GetLong(PARALLEL_GROUP, PARALLEL_NAME_MESSAGETIMEOUT);
    return TRUE;
}

// =================================================================================== 

void
TCommunication::ActivateTags(TMessageTag ActivateMask)

{
    //    Print("activate tags %ld", ActivateMask);
    ActiveTagMask |= ActivateMask;
    TagMaskHasChanged = TRUE;
}

// =================================================================================== 

bool
TCommunication::CancelRecipient(TMessageRecipient *Recipient)

{
    int  CRep;
    bool Found;
    
    CRep = 0;
    Found = FALSE;
    while (CRep < NrOfRecipients AND !Found)
        if (Recipients[CRep] == Recipient)
            Found = TRUE;
        else
            CRep++;
    if (!Found)
    {
        Error("CancelRecipient: cannot find recipient");
        return FALSE;
    }
    Print(DM_MSG_CONTROL, "cancel recipient %d handling tag mask %lX", 
          CRep, TagMasks[CRep]);
    NrOfRecipients--;
    if (CRep != NrOfRecipients)
    {
        memmove((void *) (TagMasks + CRep), (void *) (TagMasks + CRep + 1),
                sizeof(TMessageTag) * (NrOfRecipients - CRep));
        memmove((void *) (Recipients + CRep), (void *) (Recipients + CRep + 1),
                sizeof(TMessageRecipient *) * (NrOfRecipients - CRep));
    }
    return TRUE;
}

// =================================================================================== 

bool
TCommunication::CheckCapabilities(long RequiredCapabilities) const

{
    int i;

    for (i = 0; i < NrOfControllers; i++)
        if ((Controllers[i]->GetCapabilities() & RequiredCapabilities) == RequiredCapabilities)
            return TRUE;
    return FALSE;
}

// =================================================================================== 

bool
TCommunication::CheckReduceParameters(TBaseOperation BaseOperation, 
                                      TDestination &Destination,
                                      TCPNID &Root, TCommunicationControl *&Control)

{
    if (!GetControl(CC_LINEAR_BROADCAST, Control))
        return FALSE;
    if ((BaseOperation < 0) OR (BaseOperation >= BaseOperationCount))
    {
        Error("ReduceLinearArray: invalid base operation %d", BaseOperation);
        return FALSE;
    }
    if (Destination.GetNrOfRecipients() != 1)
    {
        Error("CheckReduceParameters: number of root recipients != 1");
        return FALSE;
    }
    Root = Destination.GetRecipient();
    return TRUE;
}

// =================================================================================== 

bool
TCommunication::Commit(TBaseMessage *&aMessage)

{
    if (aMessage == 0)
    {
        Warning("Commit: null message");
        return TRUE;
    }
    SendTimer->Select();
    if (aMessage->GetFlags() & MSG_FLAG_AUTO)
        if (!aMessage->EndSend())
        {
            Error("Commit: error in EndSend");
            return FALSE;
        }
    SendTimer->Unselect();
    debug(Print(DM_MSG_CONTROL, "commit (%s)", (char *) aMessage->GetObjectName()));
    if (aMessage->GetFlags() & MSG_FLAG_DELETE)
        DeleteObject(aMessage);
    return TRUE;
}

// =================================================================================== 

bool
TCommunication::Deactivate(void)

{
    int i;

    Print(DM_TIME, "idle:%.1f [s], send:%.1f [s], receive: %.1f [s], all:%.1f [s]",
          IdleTimer->GetElapsedTime(), SendTimer->GetElapsedTime(), 
          ReceiveTimer->GetElapsedTime(), CommunicationTimer->GetElapsedTime());
    DeleteObject(IdleTimer);
    DeleteObject(SendTimer);
    DeleteObject(ReceiveTimer);
    DeleteObject(CommunicationTimer);
    for (i = NrOfControllers - 1; i >= 0; i--)
        DeleteObject(Controllers[i]);
    if (PendingMessages)
    {
        DumpPendingMessages();
        PendingMessages->DeleteAll(TRUE);
        DeleteObject(PendingMessages);
    }
    Communication = 0;
    return TRUE;
}

// =================================================================================== 

void
TCommunication::DeactivateTags(TMessageTag DeactivateMask)

{
    //    Print("deactivate tags %ld", DeactivateMask);
    ActiveTagMask &= ~DeactivateMask;
    TagMaskHasChanged = TRUE;
}

/* =================================================================================== */

bool
TCommunication::DeliverMessage(TBaseMessage *aMessage, bool &ExitLoop, bool &Success)

{
    int i;

    for (i = NrOfRecipients - 1; i >= 0; i--)
        if (TagMasks[i] & aMessage->GetTag())
        {
            debug(Print(DM_MSG_CONTROL, "deliver message (%s) to recipient #%d",
                        (char *) aMessage->GetInstanceName(), i));

            PendingMessages->Delete(aMessage);
            if (!aMessage->Select())
            {
                Error("DeliverMessage: Select failed");
                return FALSE;
            }
            if (!Recipients[i]->HandleMessage(aMessage, ExitLoop))
            {
                Error("DeliverMessage: error on return from message recipient");
                return FALSE;
            }
            Success = TRUE;
            return TRUE;
        }
    Warning("%s could not be delivered in loop", (char *) aMessage->GetObjectName());
    return TRUE;
}

// =================================================================================== 

bool
TCommunication::DestroyMessage(TBaseMessage *&aMessage) const

{
    if (aMessage)
    {
        if (!aMessage->EndReceive())
        {
            Error("DestroyMessage: error in EndReceive");
            return FALSE;
        }
        DeleteObject(aMessage);
    }
    return TRUE;
}

// =================================================================================== 

void
TCommunication::DumpPendingMessages(void)

{
    if (PendingMessages->GetNrOfElements())
    {
        TBaseMessageEnum CMessage;

        Print("%ld undelivered messages:", PendingMessages->GetNrOfElements());
        CMessage.WalksForwardThrough(PendingMessages);
        while (CMessage())
        {
            Print(CMessage->GetInstanceName());
            CMessage++;
        }
    }
}

// =================================================================================== 

bool
TCommunication::FillMessageList(bool Block)

{
    int  i;
    int  Count;

    Count = 0;
    do
    {
        // the first loop does a non-blocking receive of all pending messages

        debug(Print(DM_MSG_CONTROL, "filling queue with non-blocking messages"));
        i = 0;
        while (i < NrOfControllers AND Count < MAX_PREFETCH)
        {
            while (Controllers[i]->MessageIsPending() AND Count < MAX_PREFETCH)
            {
                TBaseMessage *Message = 0;
                
                ReceiveTimer->Select();
                if (!Controllers[i]->ReceiveMessage(Message, 0))
                {
                    Error("FillMessageList: while receiving non-blocking message");
                    return FALSE;
                }
                ReceiveTimer->Unselect();
                if (!Message->Deselect())
                {
                            Error("FillMessageList: cannot deselect");
                            return FALSE;
                }
                PendingMessages->Append(Message);
                Count++;
            }
            i++;
        }
        
        // the second loop does a blocking receive for one message

        if (Block AND Count == 0)
        {
            debug(Print(DM_MSG_CONTROL, "filling queue with a blocking message"));
            i = 0;
            while (i < NrOfControllers AND Count == 0)
            {
                if (Controllers[i]->IsBlockable())
                {
                    TBaseMessage *Message = 0;

                    IdleTimer->Select();
                    if (!Controllers[i]->ReceiveMessage(Message, MessageTimeOut))
                    {
                        Error("FillMessageList: while receiving blocking message");
                        return FALSE;
                    }
                    IdleTimer->Unselect();
                    if (Message)
                    {
                        if (!Message->Deselect())
                        {
                            Error("FillMessageList: cannot deselect");
                            return FALSE;
                        }
                        PendingMessages->Append(Message);
                        Count++;
                    }
                }
                i++;
            }
        }  
    }
    while (Block AND Count == 0);
    return TRUE;
}

// =================================================================================== 

TTimer *
TCommunication::GetIdleTimer(void)

{
    return IdleTimer;
}

// =================================================================================== 

bool
TCommunication::GetControl(long Capabilities, TCommunicationControl *&Control)

{
    int i;

    for (i = 0; i < NrOfControllers; i++)
        if ((Controllers[i]->GetCapabilities() & Capabilities) == Capabilities)
        {
            Control = Controllers[i];
            return TRUE;
        }
    Error("cannot find communication control with capabilities %ld", Capabilities);
    return FALSE;
}

// =================================================================================== 

bool
TCommunication::GetNewMessage(TBaseMessage *&Message, TDestination theDestination, 
                              TMessageTag theTag, 
                              TMessageType theType, TMessageFlags theFlags)

{
    TCommunicationControl *Control;
    TDestination          *NewDest;
    TMessageID             MessageID;

    if ((NewDest = new TDestination(theDestination)) == 0)
    {
        Error("GetNewMessage: out of heap");
        return 0;
    }
    if (!GetControl(CC_SEND_MESSAGE, Control))
        return FALSE;
    MessageID = NextMessageID++;
    if ((Message = Control->GetNewMessage(MessageID, Parallel->GetLocalCPNID(),
                                          NewDest, theTag, theType, theFlags)) == 0)
    {
        Error("cannot create new message");
        return FALSE;
    }
    if (theFlags & MSG_FLAG_AUTO)
        if (!Message->BeginSend())
        {
            Error("GetNewMessage: error in automatic call to BeginSend");
            DeleteObject(Message);
            return FALSE;
        }
    return TRUE;
}

// =================================================================================== 

TString
TCommunication::GetClassName(void) const

{
    return "TCommunication";
}

// =================================================================================== 

TString
TCommunication::GetInstanceName(void) const

{
    char Line[128];

    sprintf(Line, "Tags:%lX", ActiveTagMask);
    return Line;
}

// =================================================================================== 

void
TCommunication::_Init(void)

{
    int i;

    ActiveTagMask = TAG_MASK_ALL;
    TagMaskHasChanged = FALSE;
    NrOfRecipients = 0;
    NrOfControllers = 0;
    NextMessageID = 0;
    for (i = 0; i < MAX_CONTROLLERS; i++)
        Controllers[i] = 0;
    PendingMessages = new TBaseMessageList;
    Controllers[NrOfControllers++] = new TMemoryMessageControl;
#ifdef USE_PVM
    Controllers[NrOfControllers++] = new TPVMMessageControl;
#elif USE_MPI
    Controllers[NrOfControllers++] = new TMPIMessageControl;
#endif
    CommunicationTimer = new TTimer(Environment->GetOverallTimer());
    SendTimer = new TTimer(CommunicationTimer);
    ReceiveTimer = new TTimer(CommunicationTimer);
    IdleTimer = new TTimer(CommunicationTimer);
}

/* =================================================================================== */

bool
TCommunication::MessageLoop(bool Block, TCPNID Sender)

{
    TBaseMessageEnum  CMessage;
    bool              ExitLoop = FALSE;
    int               NrOfMessages;

    NrOfMessages = 0;
    while (!ExitLoop)
    {
        Print(DM_MSG_CONTROL, 
              "MessageLoop: waiting for messages with tag %lX from %lX (in queue:%ld)", 
              ActiveTagMask, Sender, PendingMessages->GetNrOfElements());
        CMessage.WalksForwardThrough(PendingMessages);
        while (CMessage() AND !ExitLoop)
        {
            bool          Success = FALSE;
            TBaseMessage *Message = CMessage;
                
            if ((CMessage->GetTag() & ActiveTagMask) > 0 AND
                (CMessage->GetSender() == Sender OR Sender == SOURCE_ALL))
            {
                if (!DeliverMessage(Message, ExitLoop, Success))
                {
                    Error("while delivering message");
                    return FALSE;
                }               
                //              if (PendingMessages->GetCurrentEnum() != &CMessage)
                //                  Warning("possible SEGV caught!");
                //              PendingMessages->SetCurrentEnum(&CMessage); // just to make sure
                if (ExitLoop)
                    debug(Print(DM_MSG_CONTROL, 
                                "MessageLoop: previous message caused exit of loop"));
            }
            CMessage++;
            if (Success)
            {
                PendingMessages->Delete(Message);
                if (!DestroyMessage(Message))
                {
                    Error("MessageLoop: cannot destroy message");
                    return FALSE;
                }
                NrOfMessages++;
            }
        }
        if (!ExitLoop)
            if (!FillMessageList(Block))
            {
                Error("while filling message queue from %lX", Sender);
                return FALSE;
            }
    }
    return TRUE;
}

// =================================================================================== 

bool
TCommunication::ProcessPendingMessages(void)

{
    TBaseMessageEnum CMessage;
    
    if (!FillMessageList(FALSE))
    {
        Error("while filling message queue");
        return FALSE;
    }
    CMessage.WalksForwardThrough(PendingMessages);
    while (CMessage())
    {
        bool          Success = FALSE;
        TBaseMessage *Message = CMessage;

        if ((CMessage->GetTag() & ActiveTagMask) > 0)
        {
            bool ExitLoop = FALSE;
            
            if (!DeliverMessage(Message, ExitLoop, Success))
            {
                Error("while delivering message");
                return FALSE;
            }
            //      if (PendingMessages->GetCurrentEnum() != &CMessage)
            //          Warning("possible SEGV caught!");
            //      PendingMessages->SetCurrentEnum(&CMessage); // just to make sure
        }
        CMessage++;
        if (Success)
        {
            PendingMessages->Delete(Message);
            if (!DestroyMessage(Message))
            {
                Error("ProcessPendingMessages: cannot destroy message");
                return FALSE;
            }
        }
    }
    return TRUE;
}

// =================================================================================== 

// Here we use TDestination for the sender, which is rather awkward, but it works.

bool
TCommunication::ReceiveMessage(TBaseMessage *&aMessage, TDestination theSender,
                               TMessageTag theTag, TMessageType theType)

{
    TBaseMessageEnum  CMessage;
    bool              ExitLoop = FALSE;

    while (!ExitLoop)
    {
        debug(Print(DM_MSG_CONTROL, "ReceiveMessage: waiting for messages from %s "
                    "tag %lX type %ld (in queue:%ld)", 
                    (char *) theSender.GetInstanceName(), theTag, theType, 
                    PendingMessages->GetNrOfElements()));
        CMessage.WalksForwardThrough(PendingMessages);
        while (CMessage() AND !ExitLoop)
        {
            if (theSender.IsRecipient(CMessage->GetSender()) AND
                CMessage->GetTag() == theTag AND
                CMessage->GetType() == theType)
            {
                aMessage = CMessage;
                ExitLoop = TRUE;
            }
            else
                CMessage++;
        }
        if (!ExitLoop)
        {
            /*
            // <<<<< TEMP
            CMessage.WalksForwardThrough(PendingMessages);
            while (CMessage())
            {
                Print(DM_MSG_CONTROL, "in queue: %s", (char *) CMessage->GetObjectName());
                CMessage++;
            }
            // >>>>> TEMP
            */
            if (!FillMessageList(TRUE))
            {
                Error("while filling message queue");
                return FALSE;
            }
        }
    }
    if (!aMessage->Select())
    {
        Error("ReceiveMessage: Select failed");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TCommunication::RegisterRecipient(TMessageTag theTagMask, TMessageRecipient *theRecipient)

{
    if (NrOfRecipients == MAX_RECIPIENTS)
    {
        Error("maximum number of recipients (%ld) reached", MAX_RECIPIENTS);
        return FALSE;
    }
    TagMasks[NrOfRecipients] = theTagMask;
    debug(Print(DM_MSG_CONTROL, "register recipient %d handling tag mask %lX", 
                NrOfRecipients, theTagMask));
    Recipients[NrOfRecipients++] = theRecipient;
    return TRUE;
}

// =================================================================================== 

bool
TCommunication::SendEvent(TEventID EventID, TDestination Destination, TMessageTag Tag, 
                          TMessageType Type, TMessageFlags MsgFlags)

{
    TBaseMessage *Message = 0;

    if (!Communication->GetNewMessage(Message, Destination, Tag, Type, MsgFlags))
    {
        Error("SendEvent: cannot get new message");
        return FALSE;
    }
    Message->Event(EventID);
    if (!Communication->Commit(Message))
    {
        Error("SendEvent: cannot commit message");
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
