/*      file     : parbalance/TStaticBalancer.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TStaticBalancer.C,v 1.9 2005/03/25 11:06:49 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <TOrthogonalBisection.h>

#include <TCPNInfo.h>
#include <TTopology.h>

#include <ParBalanceConfig.h>
#include <TConfiguration.h>
#include <TEnvironment.h>
#include <TNode.h>
#include <TEdge.h>

#include "TStaticBalancer.h"

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TStaticBalancer::TStaticBalancer(TNodeList *theNodes, TInfoID theLocationInfoID,
                                 TEdgeList *theEdges, TInfoID theAdjacenceInfoID,
                                 TTopology *theTopology,
                                 bool LocalDistributionFlag)

{
    NodeLists = 0;
    EdgeLists = 0;
    Nodes = theNodes;
    Edges = theEdges;
    Topology = theTopology;
    LocalDistribution = LocalDistributionFlag;
    LocationInfoID = theLocationInfoID;
    AdjacenceInfoID = theAdjacenceInfoID;
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_BISECTIONSTAT, 
                                           BALANCER_DEFAULT_BISECTIONSTAT));
    BisectionStat = Configuration->GetLong(BALANCER_GROUP, BALANCER_NAME_BISECTIONSTAT);
}

// =================================================================================== 

TStaticBalancer::~TStaticBalancer(void)

{
    DeleteObjectArray(NodeLists);
    DeleteObjectArray(EdgeLists);
}

// =================================================================================== 

TEdgeList *
TStaticBalancer::GetEdgesOnCPN(int CPNIndex)

{
    if (CPNIndex < 0 OR CPNIndex >= Topology->GetNrOfCPN())
    {
        Error("GetEdgesOnCPN: invalid CPN index %d", CPNIndex);
        return 0;
    }
    return &EdgeLists[CPNIndex];
}
                                 
// =================================================================================== 

TNodeList *
TStaticBalancer::GetNodesOnCPN(int CPNIndex)

{
    if (CPNIndex < 0 OR CPNIndex >= Topology->GetNrOfCPN())
    {
        Error("GetEdgesOnCPN: invalid CPN index %d", CPNIndex);
        return 0;
    }
    return &NodeLists[CPNIndex];
}
                                 
// =================================================================================== 

bool
TStaticBalancer::OrthogonalBisection(void)

{
    TOrthogonalBisection  OB;
    TNodeEnum             CNode;
    TEdgeEnum             CEdge;
    int                   CPNIndex;

    if (!OB.Allocate((int) Nodes->GetNrOfElements(), (int) Edges->GetNrOfElements(),
                     Topology->GetNrOfCPN()))
    {
        Error("OrthogonalBisection: cannot allocate");
        return FALSE;
    }
    CNode.WalksForwardThrough(Nodes);
    while (CNode())
    {
        OB.PutNode(CNode->GetParallelID(), 
                   CNode->Loc(LocationInfoID).x(), CNode->Loc(LocationInfoID).y(),
                   CNode->GetLoad(0, INVALID, LocationInfoID));
        CNode++;
    }
    CEdge.WalksForwardThrough(Edges);
    while (CEdge())
    {
        OB.PutEdge(CEdge->GetParallelID(),
                   CEdge->GetFromNodeID(AdjacenceInfoID),
                   CEdge->GetToNodeID(AdjacenceInfoID),
                   CEdge->GetLoad(0, AdjacenceInfoID, LocationInfoID),
                   !CEdge->IsDistributable());
        CEdge++;
    }
    for (CPNIndex = 0; CPNIndex < Topology->GetNrOfCPN(); CPNIndex++)
        OB.PutCPN(Topology->GetCPNInfoByRank(CPNIndex)->Performance);
    if (!(OB.Prepare() AND
          OB.Bisection(Topology->GetNrOfCPN(), BF_BEST_OF_TWO)))
    {
        Error("OrthogonalBisection: cannot bisect");
        return FALSE;
    }
    if (BisectionStat)
    {
        TString  Filename;
        FILE    *Ascii;

        Filename = Environment->ReplacePathTokens("%R.obs");
        if ((Ascii = fopen(Filename, "w")) == 0)
        {
            Error("OrthogonalBisection: cannot open stat file %s", (char *) Filename);
            return FALSE;
        }
        OB.WriteStatHeader(Ascii);
        OB.WriteStatLine(Ascii);
        fclose(Ascii);
    }
    NodeLists = new TNodeList[Topology->GetNrOfCPN()];
    EdgeLists = new TEdgeList[Topology->GetNrOfCPN()];
    CNode.WalksForwardThrough(Nodes);
    while (CNode())
    {
        int CPNIndex;

        if ((CPNIndex = OB.GetCPNIndexByNodeID(CNode->GetParallelID())) == NIL)
        {
            Error("cannot find CPN for node %ld", CNode->GetParallelID());
            return FALSE;
        }
        NodeLists[CPNIndex].Append(CNode);
        CNode++;
    }
    CEdge.WalksForwardThrough(Edges);
    while (CEdge())
    {
        int CPNIndex;

        if ((CPNIndex = OB.GetCPNIndexByNodeID(CEdge->GetFromNodeID())) == NIL)
        {
            Error("cannot find CPN for FromNode %ld", CEdge->GetFromNodeID());
            return FALSE;
        }
        EdgeLists[CPNIndex].Append(CEdge);
        if ((CPNIndex = OB.GetCPNIndexByNodeID(CEdge->GetToNodeID())) == NIL)
        {
            Error("cannot find CPN for ToNode %ld", CEdge->GetToNodeID());
            return FALSE;
        }
        EdgeLists[CPNIndex].Append(CEdge);
        CEdge++;
    }
    return TRUE;
}

// =================================================================================== 
#undef ParentClass


