/*      file     : partools/TContainerObject.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TContainerObject.C,v 1.16 2005/05/01 14:57:58 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <stdio.h>

#include <TBaseStream.h>
#include <TObjectManager.h>

#include "TTransferInfo.h"
#include "TClassInfo.h"
#include "TObjectTimer.h"

#include "TContainerObject.h"

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#  include "TContainerObject.i.C"
#endif

// =================================================================================== 
#define ParentClass TParallelListObject
// =================================================================================== 

TContainerObject::TContainerObject(TClassInfo *theClassInfo, TParallelID theParallelID) :

ParentClass(theClassInfo, theParallelID)

{
    int i;

    Infos = new TParallelObject * [GetClassInfo()->GetNrOfInfoClasses()];
    for (i = 0; i < GetClassInfo()->GetNrOfInfoClasses(); i++)
        Infos[i] = 0;
    LoadableMask = 0;
    RequiredMask = 0;
    NrOfInstances = 0;
    LocalMask = 0;
    CPNID = INVALID;
}

// =================================================================================== 

TContainerObject::~TContainerObject(void)

{
    if (Infos)
    {
        int CInfo;
    
        for (CInfo = 0; CInfo < GetClassInfo()->GetNrOfInfoClasses(); CInfo++)
            DeleteObject(Infos[CInfo]);
        DeleteObjectArray(Infos);
    }
}

// =================================================================================== 

bool
TContainerObject::Decode(TBaseStream *Stream, TTransferInfo *TransferInfo)

{
    if (!ParentClass::Decode(Stream, TransferInfo))
        return FALSE;
    switch (TransferInfo->TransferMode)
    {
    case TM_ConstData:
        
        return (Stream->Decode(CPNID) AND
                Stream->Decode(LoadableMask) AND 
                Stream->Decode(RequiredMask));

    case TM_MigrationInfo:
        
        return Stream->Decode(CPNID);

    default:
        ;
    }
    return TRUE;
}

// =================================================================================== 

bool
TContainerObject::DetermineCPNID(void)

{
    return TRUE;
}

// =================================================================================== 

void
TContainerObject::DisposeNonLocalInfos(void)

{
    int       CInfo;
    TInfoMask Mask;
    
    Mask = 1;
    for (CInfo = 0; CInfo < GetClassInfo()->GetNrOfInfoClasses(); CInfo++)
    {
        if ((LocalMask & Mask) == 0)
            DeleteObject(Infos[CInfo]);
        Mask = Mask << 1;
    }
}

// =================================================================================== 

bool
TContainerObject::DisposeNonLocalPortion(TCPNID)

{
    return TRUE;
}

// =================================================================================== 

bool
TContainerObject::Encode(TBaseStream *Stream, TTransferInfo *TransferInfo) const
    
{
    if (!ParentClass::Encode(Stream, TransferInfo))
        return FALSE;
    switch (TransferInfo->TransferMode)
    {
    case TM_ConstData:

        return (Stream->Encode(CPNID) AND
                Stream->Encode(LoadableMask) AND
                Stream->Encode(RequiredMask));

    case TM_MigrationInfo:

        return Stream->Encode(CPNID);
        
    default:
        ;
    }
    return TRUE;
}

// =================================================================================== 

double
TContainerObject::GetActiveFraction(TCPNID) const

{
    Warning("GetActiveFraction: don't know! Assuming 1.0.");
    return 1.0;
}

// =================================================================================== 

bool
TContainerObject::GetBoundaries(TTimeStep &TimeStep, TDLListPrimitive *Boundaries)

{
    int CInfo;

    for (CInfo = 0; CInfo < GetClassInfo()->GetNrOfInfoClasses(); CInfo++)
        if (Infos[CInfo])
            if (!Infos[CInfo]->GetBoundaries(TimeStep, Boundaries))
            {
                Error("GetBoundaries: cannot get boundaries from %s",
                      (char *) Infos[CInfo]->GetObjectName());
                return FALSE;
            }
    return TRUE;
}

// =================================================================================== 

TCPNID
TContainerObject::GetCPNID(void) const
    
{
    return CPNID;
}

// =================================================================================== 

bool
TContainerObject::GetDependantContainers(int &NrOfRefs, TContainerReference *) 

{
    NrOfRefs = 0;
    return TRUE;
}

// =================================================================================== 

TParallelObject *
TContainerObject::GetInfo(TInfoID theInfoID, bool Create, bool Load) const

{
    if (theInfoID < 0 OR theInfoID >= GetClassInfo()->GetNrOfInfoClasses())
    {
        Error("GetInfo: invalid info ID %d", theInfoID);
        return 0;
    }
    if (Infos[theInfoID] == 0 AND Create)
    {
        TClassInfo *InfoClassInfo = GetClassInfo()->GetInfoClassInfo(theInfoID);

        if ((Infos[theInfoID] = 
             ObjectManager->CreateObject(InfoClassInfo, GetParallelID())) == 0)
        {
            Error("GetInfo: cannot create info");
            return 0;
        }
        if (Load)
        {
            bool Loaded = FALSE;

            if (!ObjectManager->LoadObject(InfoClassInfo, Infos[theInfoID], Loaded))
                Error("GetInfo: error while loading %s", 
                      (char *) Infos[theInfoID]->GetObjectName());
            if (!Loaded)
                Warning("GetInfo: cannot load %s", 
                        (char *) Infos[theInfoID]->GetObjectName());
        }
    }
    return Infos[theInfoID];
}

// =================================================================================== 

TString
TContainerObject::GetInstanceName(void) const

{
    char Line[256];

    sprintf(Line, "(%s) CPNID:%lX Loadable:%lld Local:%lld NoI:%d", 
            (char *) ParentClass::GetInstanceName(), 
            CPNID, LoadableMask, LocalMask, NrOfInstances);
    return Line;
}

// =================================================================================== 

double
TContainerObject::GetLoad(TCPNID LocalCPNID, TInfoID AdjacenceInfoID, TInfoID LocationInfoID)

{
    double Load;
    int    CInfo;

    Load = 0.0;
    for (CInfo = 0; CInfo < GetClassInfo()->GetNrOfInfoClasses(); CInfo++)
        if (Infos[CInfo])
            Load += Infos[CInfo]->GetLoad(LocalCPNID, AdjacenceInfoID, LocationInfoID);
    return Load;
}

// =================================================================================== 

bool
TContainerObject::GetRequiredContainers(int &NrOfRefs, TContainerReference *) 

{
    NrOfRefs = 0;
    return TRUE;
}

// =================================================================================== 

bool
TContainerObject::IsDistributable(void) const

{
    int              CInfo;
    TParallelObject *Object;

    for (CInfo = 0; CInfo < GetClassInfo()->GetNrOfInfoClasses(); CInfo++)
        if ((Object = GetInfo(CInfo)))
            if (!Object->IsDistributable())
                return FALSE;
    return TRUE;
}

// =================================================================================== 

bool
TContainerObject::IsLocal(TCPNID theCPNID) const

{
    if (CPNID == INVALID)
        Warning("IsLocal: CPNID is INVALID");
    return CPNID == theCPNID;
}

// =================================================================================== 

void
TContainerObject::RemoveInstance(void)

{
    if (NrOfInstances <= 0)
    {
        Warning("RemoveInstance: no instance present");
        return;
    }
    NrOfInstances--;
    if (NrOfInstances == 0)
        DisposeNonLocalInfos();
}

// =================================================================================== 

bool
TContainerObject::StartTimer(TInfoID InfoID)

{
    TObjectTimer *Timer;

    if ((Timer = (TObjectTimer *) GetInfo(InfoID, TRUE, TRUE)) == 0)
    {
        Error("StartTimer: cannot retrieve timer info");
        return FALSE;
    }
    Timer->Start();
    return TRUE;
}

// =================================================================================== 

bool
TContainerObject::StopTimer(TInfoID InfoID, int Interval)

{
    TObjectTimer *Timer;

    if ((Timer = (TObjectTimer *) GetInfo(InfoID)) == 0)
    {
        Error("StopTimer: cannot retrieve timer info");
        return FALSE;
    }
    Timer->Stop(Interval);
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
